# to run this application start your virtual environment and execute
# uvicorn telemetry:app --reload
import os
import dotenv
from fastapi import FastAPI, Request, HTTPException
from slowapi import Limiter, _rate_limit_exceeded_handler
from slowapi.util import get_remote_address
from slowapi.errors import RateLimitExceeded
from pydantic import BaseModel
from pymongo import MongoClient
from datetime import datetime
from typing import Union
import json

# telemetry config file location
# currently fails when attempting to serve this application at port 80 or 443 as they need to be run as root/sudo
telemetry_config = os.path.join(os.path.expanduser("~"), ".telemetry.env")

# load the environment variables from the users home dire telemetry.env file
dotenv.load_dotenv(telemetry_config, override=True)


def get_database():
    # Provide the mongodb atlas url to connect python to mongodb using pymongo
    # CONNECTION_STRING = "mongodb://127.0.0.1/telemetry"
    CONNECTION_STRING = os.environ.get("MONGO_DB_ADDRESS", "127.0.0.1:27017")

    client = MongoClient(CONNECTION_STRING)

    return client["telemetry"]


# add additional collections here
collection_names = ["ezbids", "pet2bids", "telemetry", "petdeface"]
db_name = get_database()


class Item(BaseModel):
    timestamp: str = ""
    client_addr: str = ""
    description: str = ""
    content: dict = {}


app = FastAPI()

# set up limiters so that we don't get spammed to death, limiter can be disabled be setting
# the environment variable TELEMETRY_RATE_LIMITING to False
limiter = Limiter(key_func=get_remote_address)
if (
    os.environ.get("TELEMETRY_RATE_LIMITING", "") == "False"
    or os.environ.get("TELEMETRY_RATE_LIMITING", "") == ""
):
    limiter.enabled = False
if os.environ.get("TELEMETRY_RATE_LIMIT", "") == "True":
    limiter.enabled = True

app.state.limiter = limiter
app.add_exception_handler(RateLimitExceeded, _rate_limit_exceeded_handler)


# note the route decorator must be above the limit decorator, not below it
@app.get("/")
@limiter.limit("5/minute")
async def root(request: Request):
    return {"message": "This is a very basic api"}


# use fastapi to accept a json via a post request
@app.post("/{url_string}/")
@app.post("/telemetry/")
@limiter.limit("5/minute")
async def create_item(request: Request, url_string: str = "telemetry"):
    # collect the ip address of the incoming request
    # and update the item.dict with the request ip address and host name
    try:
        data = await request.json()
    except:
        raise HTTPException(status_code=400, detail="Invalid JSON")
    description = data.get("description", "")
    item = Item()
    item.timestamp = datetime.now().isoformat()
    item.client_addr = request.client.host
    item.content = data
    item.description = description

    if url_string in collection_names:
        collection = db_name[url_string]
    else:
        # return a message to the user that the collection does not exist with fastapi's exception handling
        raise HTTPException(
            status_code=404, detail=f"You've reached a dead end at {url_string}"
        )

    # insert the json into the mongodb collection
    collection.insert_one(item.dict())
    return item


@app.get("/check/{collection}/{number_of_records}")
@app.get(
    "/check/{collection}/"
)  # if no number is provided default to 0 and get everything in the collection
async def check(collection: str = "telemetry", number_of_records: int = 0):
    if collection not in collection_names:
        raise HTTPException(
            status_code=404, detail=f"Collection {collection} does not exist"
        )

    collection_name = db_name[collection]
    if number_of_records < 0:
        if abs(number_of_records) > collection_name.count_documents({}):
            number_of_records = collection_name.count_documents({})
        cursor = list(
            collection_name.find({}).skip(
                collection_name.count_documents({}) - abs(number_of_records)
            )
        )
    elif number_of_records == 0:
        cursor = list(collection_name.find({}))
    else:
        cursor = list(collection_name.find({}).limit(number_of_records))

    for i, item in enumerate(cursor):
        # this is poor man's serialization b/c the bson.objectid.ObjectId is not serializable
        # with any built in methods
        item["_id"] = str(item["_id"])
        cursor[i] = item
    return cursor
